/*
* SpamFilter Module for UnrealIRCD
* Filters all private messages for spam and surpresses it
* by decoder, property of CyberArmy IRC Networks
* Basic code taken from modules of Syzop, visit www.unrealircd.com to get those
* Thnx to keoki, chawmp and booto for helping me with some stuff :D
* $Id: m_spam.c,v 1.5 2004/12/01 01:18:29 tirc Exp $
* Modded for QuakeIRCd - Tiburon
*/

#include "config.h"
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <io.h>
#endif
#include <fcntl.h>
#include "h.h"
#ifdef STRIPBADWORDS
#include "badwords.h"
#endif
#ifdef _WIN32
#include "version.h"
#endif
#ifndef DYNAMIC_LINKING
ModuleHeader m_spamfilter_Header
#else
#define m_spamfilter_Header Mod_Header
ModuleHeader Mod_Header
#endif
    = {
    "m_spam",             /* Name of module */
    "v1.5.0",                   /* Version */
    "private message spam filter",      /* Short description of module */
    "CyberArmy IRC Networks",
    NULL
};

#define SpamFree(x) if (x) { free(x); x = NULL; }

#define MSG_SPAMFILTER	"SPAM"
#define TOK_SPAMFILTER 	"SP"
#define CONF_FILE_NAME	"conf/spam.conf"
long UMODE_SPAMFILTER = 0;

ModuleInfo spamfilterModInfo;
FILE *config_file;
char *loadedconf = NULL;

static Hook *CheckPmMsg;


DLLFUNC char *spamfilter_checkpmmsg(aClient *, aClient *, aClient *,
                                    char *, int);
DLLFUNC int m_spamfilter(aClient * cptr, aClient * sptr, int parc,
                         char *parv[]);
const char *checkbadwords(const char *str);
static const char *our_strcasestr(const char *, char *);
size_t get_config_size(void);
char *stripconfig(char *str);
int spamfilter_del(aClient * cptr, aClient * sptr, int parc, char *parv[]);
int spamfilter_add(aClient * cptr, aClient * sptr, int parc, char *parv[]);
int spamfilter_list(aClient * cptr, aClient * sptr, int parc,
                    char *parv[]);
int spamfilter_help(aClient * cptr, aClient * sptr, int parc,
                    char *parv[]);
int spamfilter_version(aClient * cptr, aClient * sptr, int parc,
                       char *parv[]);
int spamfilter_rehash(aClient * cptr, aClient * sptr, int parc,
                      char *parv[]);
int read_configfile(void);

/* The purpose of these ifdefs, are that we can "static" link the ircd if we
* want to
*/

/* This is called on module init, before Server Ready */
#ifdef DYNAMIC_LINKING
DLLFUNC int Mod_Init(ModuleInfo * modinfo)
#else
int m_spamfilter_Init(ModuleInfo * modinfo)
#endif
{
    /* Add a umode f, the filter will only apply if the user is +f, good for opers/spam hunters so they see spam */
    UMODE_SPAMFILTER = umode_gget('f');

    /* TODO: test if umode_gget failed? */
    add_Command(MSG_SPAMFILTER, TOK_SPAMFILTER, m_spamfilter, MAXPARA);

    /* Hooking */
    //      bcopy(modinfo,&spamfilterModInfo,modinfo->size); /*segfaults the ircd?*/
    CheckPmMsg =
        HookAddPCharEx(spamfilterModInfo.handle, HOOKTYPE_USERMSG,
                       spamfilter_checkpmmsg);
}

/* Is first run when server is 100% ready */
#ifdef DYNAMIC_LINKING
DLLFUNC int Mod_Load(int module_load)
#else
int m_spamfilter_Load(int module_load)
#endif
{
    if (read_configfile()) {
        sendto_realops
            ("Successfully loaded spam filter module configuration on %s",
             me.name);
        return MOD_SUCCESS;
    } else {
        sendto_realops
            ("Failed to load the spam filter module configuration on %s, exiting...",
             me.name);
        return -1;
    }
}


/* Called when module is unloaded */
#ifdef DYNAMIC_LINKING
DLLFUNC int Mod_Unload(int module_unload)
#else
int m_spamfilter_Unload(int module_unload)
#endif
{
    /*remove spamfilter command */
    if (del_Command(MSG_SPAMFILTER, TOK_SPAMFILTER, m_spamfilter) < 0) {
        sendto_realops("Failed to delete commands when unloading %s",
                       m_spamfilter_Header.name);
        return MOD_FAILED;
    }
    /* Remove umode :( */
    umode_delete('f', UMODE_SPAMFILTER);
    UMODE_SPAMFILTER = 0;

    /* And tha hook :/ */
    HookDel(CheckPmMsg);


}

static const char *our_strcasestr(const char *haystack, char *needle)
{                               /* own strcasestr function from unrealircd */
    int i;
    int nlength = strlen(needle);
    int hlength = strlen(haystack);

    if (nlength > hlength)
        return NULL;
    if (hlength <= 0)
        return NULL;
    if (nlength <= 0)
        return haystack;
    for (i = 0; i <= (hlength - nlength); i++) {
        if (strncasecmp(haystack + i, needle, nlength) == 0)
            return haystack + i;
    }
    return NULL;                /* not found */
}

DLLFUNC char *spamfilter_checkpmmsg(aClient * cptr, aClient * sptr,
                                    aClient * acptr, char *text,
                                    int notice)
{
    if ((acptr->umodes & UMODE_SPAMFILTER))
	{
	if (!text) //wtf we supposed to do here if thers no text?
		return text;
	if (!checkbadwords(StripControlCodes(text)))
		return text;
        sendto_one(sptr,
                   ":%s NOTICE %s : Cannot send: %s is not accepting spam messages (+f)",
                   me.name, sptr->name, acptr->name);
        return NULL;            /* ignored >:) */
    } else
        return text;            /*send pm to user */
}




const char *checkbadwords(const char *str)
{
    size_t size;
    char *tmp_ptr;
    char *tmp;

    if (!str)
	return NULL;

    size = strlen(loadedconf);
    tmp_ptr = malloc(size + 1);
    tmp_ptr[size] = '\0';
    memcpy(tmp_ptr, loadedconf, size);
    for (tmp = strtok(tmp_ptr, "\n"); tmp; tmp = strtok(NULL, "\n")) {
        if (our_strcasestr(str, tmp)) {
	    SpamFree(tmp_ptr);	    
            return tmp;
        }
    }
    SpamFree(tmp_ptr);
    return NULL;
}

DLLFUNC int m_spamfilter(aClient * cptr, aClient * sptr, int parc,
                         char *parv[])
{
	if ((!IsNetAdmin(sptr) && !IsCoAdmin(sptr)) && !IsULine(sptr)) {
        sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
        return -1;
    }
    if (parc < 2) {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
                   me.name, parv[0], "SPAM");
        return -1;
    }
    if (stricmp(parv[1], "list") == 0) {
        return spamfilter_list(cptr, sptr, parc, parv);
    } else if (stricmp(parv[1], "add") == 0) {
        return spamfilter_add(cptr, sptr, parc, parv);
    } else if (stricmp(parv[1], "del") == 0) {
        return spamfilter_del(cptr, sptr, parc, parv);
    } else if (stricmp(parv[1], "help") == 0) {
        return spamfilter_help(cptr, sptr, parc, parv);
    } else if (stricmp(parv[1], "version") == 0) {
        return spamfilter_version(cptr, sptr, parc, parv);
    } else if (stricmp(parv[1], "rehash") == 0) {
        return spamfilter_rehash(cptr, sptr, parc, parv);
    } else {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
                   "SPAM");
        return -1;
    }
}

int spamfilter_add(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
    if (parc < 4) {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
                   me.name, parv[0], "SPAM");
        return -1;
    }
    config_file = fopen(CONF_FILE_NAME, "a");
    if (!config_file) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    if (!fprintf(config_file, "%s\n", parv[3])) {
        sendto_realops("Failed to write to configfile %s", CONF_FILE_NAME);
        return -1;
    }
    if (fflush(config_file)) {
        sendto_realops("Failed to write/flush to configfile %s",
                       CONF_FILE_NAME);
        return -1;
    }
    fclose(config_file);
    if (stricmp(parv[2], "global") == 0) {
        sendto_serv_butone_token(cptr, sptr->name, MSG_SPAMFILTER,
                                 TOK_SPAMFILTER, "%s %s :%s", parv[1],
                                 "local", parv[3]);
        sendto_realops
            ("%s (%s!%s@%s) globally added spamline \"%s\" to %s",
             sptr->name, sptr->name, sptr->user->username,
             sptr->user->virthost, parv[3], CONF_FILE_NAME);
    } else {
        sendto_realops("%s (%s!%s@%s) added spamline \"%s\" to %s",
                       sptr->name, sptr->name, sptr->user->username,
                       sptr->user->virthost, parv[3], CONF_FILE_NAME);
    }
    return 1;
}

int spamfilter_list(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
    int word_found = 0;
    char buffer[512];

    config_file = fopen(CONF_FILE_NAME, "r");
    if (!config_file) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    sendto_one(sptr, ":%s NOTICE %s : SpamFilter list contains:", me.name,
               sptr->name);
    while (fgets(buffer, 512, config_file)) {
        word_found++;
        sendto_one(sptr, ":%s NOTICE %s : %s", me.name, sptr->name,
                   buffer);
    }
    if (!word_found) {
        sendto_one(sptr, ":%s NOTICE %s : Spamlist contains no entries!",
                   me.name, sptr->name);
        return -1;
    }
    sendto_one(sptr, ":%s NOTICE %s : End of List", me.name, sptr->name);
    return 1;
}

int spamfilter_del(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
    int word_found = 0;
    char *ptr;
    char *tmp;
    size_t size = get_config_size();
    if (parc < 4) {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
                   me.name, parv[0], "SPAM");
        return -1;
    }
    if (!size) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    config_file = fopen(CONF_FILE_NAME, "r");
    if (!config_file) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    ptr = malloc(size + 1);
    ptr[size] = '\0';
    fread(ptr, sizeof(char), size, config_file);
    rewind(config_file);
    fclose(config_file);
    config_file = fopen(CONF_FILE_NAME, "w");
    if (!config_file) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    for (tmp = strtok(ptr, "\n"); tmp; tmp = strtok(NULL, "\n")) {
        if ((stricmp(tmp, parv[3]))) {
            fprintf(config_file, "%s\n", tmp);
            fflush(config_file);
        } else {
            word_found++;
        }
    }
    if (!word_found) {
        if (stricmp(parv[2], "global") == 0) {
            sendto_serv_butone_token(cptr, sptr->name, MSG_SPAMFILTER,
                                     TOK_SPAMFILTER, "%s %s :%s", parv[1],
                                     "local", parv[3]);
            sendto_realops
                ("%s (%s!%s@%s) globally deleted spamline \"%s\" to %s",
                 sptr->name, sptr->name, sptr->user->username,
                 sptr->user->virthost, parv[3], CONF_FILE_NAME);
        }
        sendto_one(sptr,
                   ":%s NOTICE %s : Failed to delete \"%s\" from %s (Entry not found)",
                   me.name, sptr->name, parv[3], CONF_FILE_NAME);
        fclose(config_file);
        free(ptr);
        return -1;
    }
    fflush(config_file);
    fclose(config_file);
    free(ptr);
    if (stricmp(parv[2], "global") == 0) {
        sendto_serv_butone_token(cptr, sptr->name, MSG_SPAMFILTER,
                                 TOK_SPAMFILTER, "%s %s :%s", parv[1],
                                 "local", parv[3]);
        sendto_realops
            ("%s (%s!%s@%s) globally deleted spamline \"%s\" to %s",
             sptr->name, sptr->name, sptr->user->username,
             sptr->user->virthost, parv[3], CONF_FILE_NAME);
    } else {
        sendto_realops("%s (%s!%s@%s) deleted spamline \"%s\" to %s",
                       sptr->name, sptr->name, sptr->user->username,
                       sptr->user->virthost, parv[3], CONF_FILE_NAME);
    }
    return 1;
}

int spamfilter_help(aClient * cptr, aClient * sptr, int parc, char *parv[])
{
    sendto_one(sptr, ":%s NOTICE %s :Available commands are:", me.name,
               sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :ADD local :line to add - Adds a line to the spamfilter config of the local server",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :ADD global :line to add - Adds a line to the spamfilter config of all servers",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :DEL local :line to del - Deletes a line from the spamfilter config of the local server",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :DEL global :line to del - Deletes a line from the spamfilter config of all servers",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :LIST - Displays all entries of spamfilter configuration file",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :REHASH local - Locally rereads the spamfilter config file into memory.",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :REHASH global - Globally rereads the spamfilter config file into memory.",
               me.name, sptr->name);
    sendto_one(sptr, ":%s NOTICE %s :HELP - Displays this message",
               me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :VERSION - Displays version information",
               me.name, sptr->name);
    sendto_one(sptr, ":%s NOTICE %s :-", me.name, sptr->name);
    sendto_one(sptr,
               ":%s NOTICE %s :NOTE: ADD and DEL command require a REHASH for the changes to take effect",
               me.name, sptr->name);
    return 1;
}

int spamfilter_version(aClient * cptr, aClient * sptr, int parc,
                       char *parv[])
{
    sendto_one(sptr,
               ":%s NOTICE %s :QuakeIRCd SpamFilter Module %s by decoder",
               me.name, sptr->name, m_spamfilter_Header.version);
    sendto_one(sptr,
               ":%s NOTICE %s :CyberArmy IRC Networks",
               me.name, sptr->name);
    return 1;
}

int spamfilter_rehash(aClient * cptr, aClient * sptr, int parc,
                      char *parv[])
{
    if ((parv[2]) && (stricmp(parv[2], "global") == 0)) {
        sendto_serv_butone_token(cptr, sptr->name, MSG_SPAMFILTER,
                                 TOK_SPAMFILTER, "%s %s", parv[1],
                                 "local");
        sendto_realops
            ("%s is globally rehashing the spam filter module configuation from %s",
             sptr->name, me.name);
    } else {
        sendto_realops
            ("%s is rehashing the spam filter module configuation on %s",
             sptr->name, me.name);
    }
    if (read_configfile()) {
        sendto_realops
            ("Successfully rehashed spam filter module configuration on %s",
             me.name);
        return 1;
    } else {
        sendto_realops
            ("Failed to rehash the spam filter module configuration on %s",
             me.name);
        return -1;
    }
}

int read_configfile(void)
{
    size_t size = get_config_size();
    if (!size) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    config_file = fopen(CONF_FILE_NAME, "r");
    if (!config_file) {
        sendto_realops("Failed to open configfile %s", CONF_FILE_NAME);
        return -1;
    }
    loadedconf = malloc(size + 1);
    loadedconf[size] = '\0';
    fread(loadedconf, sizeof(char), size, config_file);
    rewind(config_file);
    fclose(config_file);
    return 1;
}

size_t get_config_size(void)
{
    size_t size;
    config_file = fopen(CONF_FILE_NAME, "rb");
    if (!config_file)
        return 0;
    fseek(config_file, 0L, SEEK_END);
    size = ftell(config_file);
    fclose(config_file);
    return size;
}
